@@ -1,6 +1,6 @@
 /*****************************************************************************
  * Free42 -- an HP-42S calculator simulator
- * Copyright (C) 2004-2021  Thomas Okken
+ * Copyright (C) 2004-2022  Thomas Okken
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -15,13 +15,6 @@
  * along with this program; if not, see http://www.gnu.org/licenses/.
  *****************************************************************************/
 
-// to enable debug out
-//#define WINDOWS_DEBUG
-
-#ifdef WINDOWS_DEBUG
-#include <windows.h> 
-#endif 
-
 #include <stdlib.h>
 
 #include "core_math1.h"
@@ -38,7 +31,7 @@
 #define NUM_SHADOWS 10
 
 /* Solver */
-typedef struct {
+struct solve_state {
     int version;
     char prgm_name[7];
     int prgm_length;
@@ -63,7 +56,8 @@ typedef struct {
     int shadow_length[NUM_SHADOWS];
     phloat shadow_value[NUM_SHADOWS];
     uint4 last_disp_time;
-} solve_state;
+    int prev_sp;
+};
 
 static solve_state solve;
 
@@ -72,7 +66,7 @@ static solve_state solve;
 #define ROMB_MAX 20
 
 /* Integrator */
-typedef struct {
+struct integ_state {
     int version;
     char prgm_name[7];
     int prgm_length;
@@ -95,7 +89,8 @@ typedef struct {
     phloat t, u;
     phloat prev_int;
     phloat prev_res;
-} integ_state;
+    int prev_sp;
+};
 
 static integ_state integ;
 
@@ -140,6 +135,7 @@ bool persist_math() {
         if (!write_phloat(solve.shadow_value[i])) return false;
     }
     if (!write_int4(solve.last_disp_time)) return false;
+    if (!write_int(solve.prev_sp)) return false;
 
     if (!write_int(integ.version)) return false;
     if (fwrite(integ.prgm_name, 1, 7, gfile) != 7) return false;
@@ -174,6 +170,7 @@ bool persist_math() {
     if (!write_phloat(integ.u)) return false;
     if (!write_phloat(integ.prev_int)) return false;
     if (!write_phloat(integ.prev_res)) return false;
+    if (!write_int(integ.prev_sp)) return false;
     return true;
 }
 
@@ -219,7 +216,12 @@ bool unpersist_math(int ver, bool discard) {
             if (!read_phloat(&solve.shadow_value[i])) return false;
         }
         if (!read_int4((int4 *) &solve.last_disp_time)) return false;
-        
+        if (ver >= 33) {
+            if (!read_int(&solve.prev_sp)) return false;
+        } else {
+            solve.prev_sp = -2;
+        }
+
         if (!read_int(&integ.version)) return false;
         if (fread(integ.prgm_name, 1, 7, gfile) != 7) return false;
         if (!read_int(&integ.prgm_length)) return false;
@@ -253,6 +255,11 @@ bool unpersist_math(int ver, bool discard) {
         if (!read_phloat(&integ.u)) return false;
         if (!read_phloat(&integ.prev_int)) return false;
         if (!read_phloat(&integ.prev_res)) return false;
+        if (ver >= 33) {
+            if (!read_int(&integ.prev_sp)) return false;
+        } else {
+            integ.prev_sp = -2;
+        }
     } else {
         int size;
         bool success;
@@ -303,6 +310,15 @@ void reset_math() {
     reset_integ();
 }
 
+static void clean_stack(int prev_sp) {
+    if (flags.f.big_stack && prev_sp != -2 && sp > prev_sp) {
+        int excess = sp - prev_sp;
+        for (int i = 0; i < excess; i++)
+            free_vartype(stack[sp - i]);
+        sp -= excess;
+    }
+}
+
 static void reset_solve() {
     int i;
     for (i = 0; i < NUM_SHADOWS; i++)
@@ -394,6 +410,7 @@ static int call_solve_fn(int which, int state) {
     arg.length = solve.active_prgm_length;
     for (i = 0; i < arg.length; i++)
         arg.val.text[i] = solve.active_prgm_name[i];
+    clean_stack(solve.prev_sp);
     err = docmd_gto(&arg);
     if (err != ERR_NONE) {
         free_vartype(v);
@@ -416,6 +433,7 @@ int start_solve(const char *name, int length, phloat x1, phloat x2) {
                 solve.prgm_name, solve.prgm_length);
     solve.prev_prgm = current_prgm;
     solve.prev_pc = pc;
+    solve.prev_sp = flags.f.big_stack ? sp : -2;
     if (x1 == x2) {
         if (x1 == 0) {
             x2 = 1;
@@ -447,10 +465,10 @@ int start_solve(const char *name, int length, phloat x1, phloat x2) {
     return call_solve_fn(1, 1);
 }
 
-typedef struct {
+struct message_spec {
     const char *text;
     int length;
-} message_spec;
+};
 
 #define SOLVE_ROOT          0
 #define SOLVE_SIGN_REVERSAL 1
@@ -511,8 +529,11 @@ static int finish_solve(int message) {
 
     solve.state = 0;
 
+    clean_stack(solve.prev_sp);
     v = recall_var(solve.var_name, solve.var_length);
     ((vartype_real *) v)->x = b;
+    if (flags.f.big_stack && !ensure_stack_capacity(4))
+        return ERR_INSUFFICIENT_MEMORY;
     new_x = dup_vartype(v);
     new_y = new_real(s);
     new_z = new_real(final_f);
@@ -524,14 +545,15 @@ static int finish_solve(int message) {
         free_vartype(new_t);
         return ERR_INSUFFICIENT_MEMORY;
     }
-    free_vartype(reg_x);
-    free_vartype(reg_y);
-    free_vartype(reg_z);
-    free_vartype(reg_t);
-    reg_x = new_x;
-    reg_y = new_y;
-    reg_z = new_z;
-    reg_t = new_t;
+    if (flags.f.big_stack)
+        sp += 4;
+    else
+        for (int i = 0; i < 4; i++)
+            free_vartype(stack[i]);
+    stack[sp] = new_x;
+    stack[sp - 1] = new_y;
+    stack[sp - 2] = new_z;
+    stack[sp - 3] = new_t;
 
     current_prgm = solve.prev_prgm;
     pc = solve.prev_pc;
@@ -565,25 +587,11 @@ static int finish_solve(int message) {
 
     if (print && message != SOLVE_ROOT)
         print_lines(solve_message[message].text,
-                    solve_message[message].length, 1);
+                    solve_message[message].length, true);
 
     return solve.keep_running ? ERR_NONE : ERR_STOP;
 }
 
-#ifdef WINDOWS_DEBUG
-static void printout(const phloat& p, const char* s) {
-    char buf[256];
-    buf[0] = 0;
-    if (s) {
-        strcpy(buf, s);
-        strcat(buf, " ");
-    }
-    p.bcd.asString(buf + strlen(buf));
-    strcat(buf, "\n");
-    OutputDebugString(buf);
-}
-#endif
-
 int return_to_solve(int failure, bool stop) {
     phloat f, slope, s, xnew, prev_f = solve.curr_f;
     uint4 now_time;
@@ -594,8 +602,10 @@ int return_to_solve(int failure, bool stop) {
     if (solve.state == 0)
         return ERR_INTERNAL_ERROR;
     if (!failure) {
-        if (reg_x->type == TYPE_REAL) {
-            f = ((vartype_real *) reg_x)->x;
+        if (sp == -1)
+            return ERR_TOO_FEW_ARGUMENTS;
+        if (stack[sp]->type == TYPE_REAL) {
+            f = ((vartype_real *) stack[sp])->x;
             solve.curr_f = f;
             if (f == 0)
                 return finish_solve(SOLVE_ROOT);
@@ -980,7 +990,7 @@ int return_to_solve(int failure, bool stop) {
             // extended-precision math for solver internals, or at least a
             // slightly smarter implementation of the midpoint calculation,
             // e.g. fall back on x3 = x1 + (x2 - x1) / 2 if x3 = (x1 + x2) / 2
-            // returns an incorrect result. 
+            // returns an incorrect result.
             if (solve.x3 <= solve.x1 || solve.x3 >= solve.x2) {
                 solve.which = -1;
                 return finish_solve(SOLVE_ROOT);
@@ -1038,6 +1048,7 @@ static int call_integ_fn() {
     arg.length = integ.active_prgm_length;
     for (i = 0; i < arg.length; i++)
         arg.val.text[i] = integ.active_prgm_name[i];
+    clean_stack(integ.prev_sp);
     err = docmd_gto(&arg);
     if (err != ERR_NONE) {
         free_vartype(v);
@@ -1088,6 +1099,7 @@ int start_integ(const char *name, int length) {
                 integ.prgm_name, integ.prgm_length);
     integ.prev_prgm = current_prgm;
     integ.prev_pc = pc;
+    integ.prev_sp = flags.f.big_stack ? sp : -2;
 
     integ.a = integ.llim;
     integ.b = integ.ulim - integ.llim;
@@ -1116,6 +1128,7 @@ static int finish_integ() {
     int saved_trace = flags.f.trace_print;
     integ.state = 0;
 
+    clean_stack(integ.prev_sp);
     x = new_real(integ.sum * integ.b * 0.75);
     y = new_real(integ.eps);
     if (x == NULL || y == NULL) {
@@ -1124,7 +1137,8 @@ static int finish_integ() {
         return ERR_INSUFFICIENT_MEMORY;
     }
     flags.f.trace_print = 0;
-    recall_two_results(x, y);
+    if (recall_two_results(x, y) != ERR_NONE)
+        return ERR_INSUFFICIENT_MEMORY;
     flags.f.trace_print = saved_trace;
 
     current_prgm = integ.prev_prgm;
@@ -1156,7 +1170,7 @@ static int finish_integ() {
 int return_to_integ(bool stop) {
     if (stop)
         integ.keep_running = 0;
-    
+
     switch (integ.state) {
     case 0:
         return ERR_INTERNAL_ERROR;
@@ -1178,11 +1192,13 @@ int return_to_integ(bool stop) {
         return call_integ_fn();
 
     case 2:
-        if (reg_x->type == TYPE_STRING)
+        if (sp == -1)
+            return ERR_TOO_FEW_ARGUMENTS;
+        if (stack[sp]->type == TYPE_STRING)
             return ERR_ALPHA_DATA_IS_INVALID;
-        else if (reg_x->type != TYPE_REAL)
+        else if (stack[sp]->type != TYPE_REAL)
             return ERR_INVALID_TYPE;
-        integ.sum += integ.t * ((vartype_real *) reg_x)->x;
+        integ.sum += integ.t * ((vartype_real *) stack[sp])->x;
         integ.p += integ.h;
         if (++integ.i < integ.nsteps)
             goto loop2;
@@ -1220,7 +1236,7 @@ int return_to_integ(bool stop) {
 
         if (++integ.n >= ROMB_MAX)
             return finish_integ(); // too many
-        
+
         goto loop1;
     default:
         return ERR_INTERNAL_ERROR;
