@@ -1,6 +1,6 @@
 /*****************************************************************************
  * Free42 -- an HP-42S calculator simulator
- * Copyright (C) 2004-2021  Thomas Okken
+ * Copyright (C) 2004-2022  Thomas Okken
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -16,6 +16,7 @@
  *****************************************************************************/
 
 #include <stdlib.h>
+#include <string.h>
 
 #include "core_globals.h"
 #include "core_helpers.h"
@@ -26,74 +27,67 @@
 // We cache vartype_real, vartype_complex, and vartype_string instances, to
 // cut down on the malloc/free overhead.
 
-typedef struct pool_real {
-    vartype_real r;
-    struct pool_real *next;
-} pool_real;
-
-static pool_real *realpool = NULL;
-
-typedef struct pool_complex {
-    vartype_complex c;
-    struct pool_complex *next;
-} pool_complex;
-
-static pool_complex *complexpool = NULL;
-
-typedef struct pool_string {
-    vartype_string s;
-    struct pool_string *next;
-} pool_string;
-
-static pool_string *stringpool = NULL;
+#define POOLSIZE 10
+static vartype_real *realpool[POOLSIZE];
+static vartype_complex *complexpool[POOLSIZE];
+static vartype_string *stringpool[POOLSIZE];
+static int realpool_size = 0;
+static int complexpool_size = 0;
+static int stringpool_size = 0;
 
 vartype *new_real(phloat value) {
-    pool_real *r;
-    if (realpool == NULL) {
-        r = (pool_real *) malloc(sizeof(pool_real));
+    vartype_real *r;
+    if (realpool_size > 0) {
+        r = realpool[--realpool_size];
+    } else {
+        r = (vartype_real *) malloc(sizeof(vartype_real));
         if (r == NULL)
             return NULL;
-        r->r.type = TYPE_REAL;
-    } else {
-        r = realpool;
-        realpool = realpool->next;
+        r->type = TYPE_REAL;
     }
-    r->r.x = value;
+    r->x = value;
     return (vartype *) r;
 }
 
 vartype *new_complex(phloat re, phloat im) {
-    pool_complex *c;
-    if (complexpool == NULL) {
-        c = (pool_complex *) malloc(sizeof(pool_complex));
+    vartype_complex *c;
+    if (complexpool_size > 0) {
+        c = complexpool[--complexpool_size];
+    } else {
+        c = (vartype_complex *) malloc(sizeof(vartype_complex));
         if (c == NULL)
             return NULL;
-        c->c.type = TYPE_COMPLEX;
-    } else {
-        c = complexpool;
-        complexpool = complexpool->next;
+        c->type = TYPE_COMPLEX;
     }
-    c->c.re = re;
-    c->c.im = im;
+    c->re = re;
+    c->im = im;
     return (vartype *) c;
 }
 
 vartype *new_string(const char *text, int length) {
-    pool_string *s;
-    if (stringpool == NULL) {
-        s = (pool_string *) malloc(sizeof(pool_string));
-        if (s == NULL)
+    char *dbuf;
+    if (length > SSLENV) {
+        dbuf = (char *) malloc(length);
+        if (dbuf == NULL)
             return NULL;
-        s->s.type = TYPE_STRING;
-    } else {
-        s = stringpool;
-        stringpool = stringpool->next;
     }
-    int i;
-    s->s.type = TYPE_STRING;
-    s->s.length = length > 6 ? 6 : length;
-    for (i = 0; i < s->s.length; i++)
-        s->s.text[i] = text[i];
+    vartype_string *s;
+    if (stringpool_size > 0) {
+        s = stringpool[--stringpool_size];
+    } else {
+        s = (vartype_string *) malloc(sizeof(vartype_string));
+        if (s == NULL) {
+            if (length > SSLENV)
+                free(dbuf);
+            return NULL;
+        }
+        s->type = TYPE_STRING;
+    }
+    s->length = length;
+    if (length > SSLENV)
+        s->t.ptr = dbuf;
+    if (text != NULL)
+        memcpy(length > SSLENV ? s->t.ptr : s->t.buf, text, length);
     return (vartype *) s;
 }
 
@@ -118,14 +112,12 @@ vartype *new_realmatrix(int4 rows, int4 columns) {
     }
     rm->array->data = (phloat *) malloc(sz * sizeof(phloat));
     if (rm->array->data == NULL) {
-        /* Oops */
         free(rm->array);
         free(rm);
         return NULL;
     }
     rm->array->is_string = (char *) malloc(sz);
     if (rm->array->is_string == NULL) {
-        /* Oops */
         free(rm->array->data);
         free(rm->array);
         free(rm);
@@ -133,8 +125,7 @@ vartype *new_realmatrix(int4 rows, int4 columns) {
     }
     for (i = 0; i < sz; i++)
         rm->array->data[i] = 0;
-    for (i = 0; i < sz; i++)
-        rm->array->is_string[i] = 0;
+    memset(rm->array->is_string, 0, sz);
     rm->array->refcount = 1;
     return (vartype *) rm;
 }
@@ -160,7 +151,6 @@ vartype *new_complexmatrix(int4 rows, int4 columns) {
     }
     cm->array->data = (phloat *) malloc(sz * sizeof(phloat));
     if (cm->array->data == NULL) {
-        /* Oops */
         free(cm->array);
         free(cm);
         return NULL;
@@ -171,27 +161,26 @@ vartype *new_complexmatrix(int4 rows, int4 columns) {
     return (vartype *) cm;
 }
 
-vartype *new_matrix_alias(vartype *m) {
-    if (m->type == TYPE_REALMATRIX) {
-        vartype_realmatrix *rm1 = (vartype_realmatrix *) m;
-        vartype_realmatrix *rm2 = (vartype_realmatrix *)
-                                        malloc(sizeof(vartype_realmatrix));
-        if (rm2 == NULL)
-            return NULL;
-        *rm2 = *rm1;
-        rm2->array->refcount++;
-        return (vartype *) rm2;
-    } else if (m->type == TYPE_COMPLEXMATRIX) {
-        vartype_complexmatrix *cm1 = (vartype_complexmatrix *) m;
-        vartype_complexmatrix *cm2 = (vartype_complexmatrix *)
-                                        malloc(sizeof(vartype_complexmatrix));
-        if (cm2 == NULL)
-            return NULL;
-        *cm2 = *cm1;
-        cm2->array->refcount++;
-        return (vartype *) cm2;
-    } else
+vartype *new_list(int4 size) {
+    vartype_list *list = (vartype_list *) malloc(sizeof(vartype_list));
+    if (list == NULL)
         return NULL;
+    list->type = TYPE_LIST;
+    list->size = size;
+    list->array = (list_data *) malloc(sizeof(list_data));
+    if (list->array == NULL) {
+        free(list);
+        return NULL;
+    }
+    list->array->data = (vartype **) malloc(size * sizeof(vartype *));
+    if (list->array->data == NULL && size != 0) {
+        free(list->array);
+        free(list);
+        return NULL;
+    }
+    memset(list->array->data, 0, size * sizeof(vartype *));
+    list->array->refcount = 1;
+    return (vartype *) list;
 }
 
 void free_vartype(vartype *v) {
@@ -199,26 +188,34 @@ void free_vartype(vartype *v) {
         return;
     switch (v->type) {
         case TYPE_REAL: {
-            pool_real *r = (pool_real *) v;
-            r->next = realpool;
-            realpool = r;
+            if (realpool_size < POOLSIZE)
+                realpool[realpool_size++] = (vartype_real *) v;
+            else
+                free(v);
             break;
         }
         case TYPE_COMPLEX: {
-            pool_complex *c = (pool_complex *) v;
-            c->next = complexpool;
-            complexpool = c;
+            if (complexpool_size < POOLSIZE)
+                complexpool[complexpool_size++] = (vartype_complex *) v;
+            else
+                free(v);
             break;
         }
         case TYPE_STRING: {
-            pool_string *s = (pool_string *) v;
-            s->next = stringpool;
-            stringpool = s;
+            vartype_string *s = (vartype_string *) v;
+            if (s->length > SSLENV)
+                free(s->t.ptr);
+            if (stringpool_size < POOLSIZE)
+                stringpool[stringpool_size++] = s;
+            else
+                free(s);
             break;
         }
         case TYPE_REALMATRIX: {
             vartype_realmatrix *rm = (vartype_realmatrix *) v;
             if (--(rm->array->refcount) == 0) {
+                int4 sz = rm->rows * rm->columns;
+                free_long_strings(rm->array->is_string, rm->array->data, sz);
                 free(rm->array->data);
                 free(rm->array->is_string);
                 free(rm->array);
@@ -235,25 +232,80 @@ void free_vartype(vartype *v) {
             free(cm);
             break;
         }
+        case TYPE_LIST: {
+            vartype_list *list = (vartype_list *) v;
+            if (--(list->array->refcount) == 0) {
+                for (int4 i = 0; i < list->size; i++)
+                    free_vartype(list->array->data[i]);
+                free(list->array->data);
+                free(list->array);
+            }
+            free(list);
+            break;
+        }
     }
 }
 
 void clean_vartype_pools() {
-    while (realpool != NULL) {
-        pool_real *r = realpool;
-        realpool = r->next;
-        free(r);
+    while (realpool_size > 0)
+        free(realpool[--realpool_size]);
+    while (complexpool_size > 0)
+        free(complexpool[--complexpool_size]);
+    while (stringpool_size > 0)
+        free(stringpool[--stringpool_size]);
+}
+
+void free_long_strings(char *is_string, phloat *data, int4 n) {
+    for (int4 i = 0; i < n; i++)
+        if (is_string[i] == 2)
+            free(*(void **) &data[i]);
+}
+
+void get_matrix_string(vartype_realmatrix *rm, int i, char **text, int4 *length) {
+    if (rm->array->is_string[i] == 1) {
+        char *t = (char *) &rm->array->data[i];
+        *text = t + 1;
+        *length = *t;
+    } else {
+        int4 *p = *(int4 **) &rm->array->data[i];
+        *text = (char *) (p + 1);
+        *length = *p;
     }
-    while (complexpool != NULL) {
-        pool_complex *c = complexpool;
-        complexpool = c->next;
-        free(c);
+}
+
+void get_matrix_string(const vartype_realmatrix *rm, int i, const char **text, int4 *length) {
+    get_matrix_string((vartype_realmatrix *) rm, i, (char **) text, length);
+}
+
+bool put_matrix_string(vartype_realmatrix *rm, int i, const char *text, int4 length) {
+    char *ptext;
+    int4 plength;
+    if (rm->array->is_string[i] != 0) {
+        get_matrix_string(rm, i, &ptext, &plength);
+        if (plength == length) {
+            memcpy(ptext, text, length);
+            return true;
+        }
     }
-    while (stringpool != NULL) {
-        pool_string *s = stringpool;
-        stringpool = s->next;
-        free(s);
+    if (length > SSLENM) {
+        int4 *p = (int4 *) malloc(length + 4);
+        if (p == NULL)
+            return false;
+        *p = length;
+        memcpy(p + 1, text, length);
+        if (rm->array->is_string[i] == 2)
+            free(*(void **) &rm->array->data[i]);
+        *(int4 **) &rm->array->data[i] = p;
+        rm->array->is_string[i] = 2;
+    } else {
+        if (rm->array->is_string[i] == 2)
+            free(*(void **) &rm->array->data[i]);
+        char *t = (char *) &rm->array->data[i];
+        t[0] = length;
+        memmove(t + 1, text, length);
+        rm->array->is_string[i] = 1;
     }
+    return true;
 }
 
 vartype *dup_vartype(const vartype *v) {
@@ -274,10 +326,7 @@ vartype *dup_vartype(const vartype *v) {
                                         malloc(sizeof(vartype_realmatrix));
             if (rm2 == NULL)
                 return NULL;
-            rm2->type = TYPE_REALMATRIX;
-            rm2->rows = rm->rows;
-            rm2->columns = rm->columns;
-            rm2->array = rm->array;
+            *rm2 = *rm;
             rm->array->refcount++;
             return (vartype *) rm2;
         }
@@ -287,16 +336,22 @@ vartype *dup_vartype(const vartype *v) {
                                         malloc(sizeof(vartype_complexmatrix));
             if (cm2 == NULL)
                 return NULL;
-            cm2->type = TYPE_COMPLEXMATRIX;
-            cm2->rows = cm->rows;
-            cm2->columns = cm->columns;
-            cm2->array = cm->array;
+            *cm2 = *cm;
             cm->array->refcount++;
             return (vartype *) cm2;
         }
         case TYPE_STRING: {
             vartype_string *s = (vartype_string *) v;
-            return new_string(s->text, s->length);
+            return new_string(s->txt(), s->length);
+        }
+        case TYPE_LIST: {
+            vartype_list *list = (vartype_list *) v;
+            vartype_list *list2 = (vartype_list *) malloc(sizeof(vartype_list));
+            if (list2 == NULL)
+                return NULL;
+            *list2 = *list;
+            list->array->refcount++;
+            return (vartype *) list2;
         }
         default:
             return NULL;
@@ -327,10 +382,25 @@ int disentangle(vartype *v) {
                     free(md);
                     return 0;
                 }
-                for (i = 0; i < sz; i++)
-                    md->data[i] = rm->array->data[i];
-                for (i = 0; i < sz; i++)
+                for (i = 0; i < sz; i++) {
                     md->is_string[i] = rm->array->is_string[i];
+                    if (md->is_string[i] == 2) {
+                        int4 *sp = *(int4 **) &rm->array->data[i];
+                        int4 len = *sp + 4;
+                        int4 *dp = (int4 *) malloc(len);
+                        if (dp == NULL) {
+                            free_long_strings(md->is_string, md->data, i);
+                            free(md->is_string);
+                            free(md->data);
+                            free(md);
+                            return 0;
+                        }
+                        memcpy(dp, sp, len);
+                        *(int4 **) &md->data[i] = dp;
+                    } else {
+                        md->data[i] = rm->array->data[i];
+                    }
+                }
                 md->refcount = 1;
                 rm->array->refcount--;
                 rm->array = md;
@@ -361,6 +431,39 @@ int disentangle(vartype *v) {
                 return 1;
             }
         }
+        case TYPE_LIST: {
+            vartype_list *list = (vartype_list *) v;
+            if (list->array->refcount == 1)
+                return 1;
+            else {
+                list_data *ld = (list_data *) malloc(sizeof(list_data));
+                if (ld == NULL)
+                    return 0;
+                ld->data = (vartype **) malloc(list->size * sizeof(vartype *));
+                if (ld->data == NULL && list->size != 0) {
+                    free(ld);
+                    return 0;
+                }
+                for (int4 i = 0; i < list->size; i++) {
+                    vartype *vv = list->array->data[i];
+                    if (vv != NULL) {
+                        vv = dup_vartype(vv);
+                        if (vv == NULL) {
+                            for (int4 j = 0; j < i; j++)
+                                free_vartype(ld->data[j]);
+                            free(ld->data);
+                            free(ld);
+                            return 0;
+                        }
+                    }
+                    ld->data[i] = vv;
+                }
+                ld->refcount = 1;
+                list->array->refcount--;
+                list->array = ld;
+                return 1;
+            }
+        }
         case TYPE_REAL:
         case TYPE_COMPLEX:
         case TYPE_STRING:
@@ -424,6 +527,14 @@ int store_var(const char *name, int namelength, vartype *value, bool local) {
         vars[varindex].level = local ? get_rtn_level() : -1;
         vars[varindex].flags = 0;
     } else if (local && vars[varindex].level < get_rtn_level()) {
+        bool must_push = false;
+        if ((matedit_mode == 1 || matedit_mode == 3)
+                && string_equals(name, namelength, matedit_name, matedit_length)) {
+            if (matedit_mode == 3)
+                return ERR_RESTRICTED_OPERATION;
+            else
+                must_push = true;
+        }
         if (vars_count == vars_capacity) {
             int nc = vars_capacity + 25;
             var_struct *nv = (var_struct *) realloc(vars, nc * sizeof(var_struct));
@@ -439,7 +550,8 @@ int store_var(const char *name, int namelength, vartype *value, bool local) {
             vars[varindex].name[i] = name[i];
         vars[varindex].level = get_rtn_level();
         vars[varindex].flags = VAR_HIDING;
-        push_indexed_matrix(name, namelength);
+        if (must_push)
+            push_indexed_matrix();
     } else {
         if (matedit_mode == 1 &&
                 string_equals(name, namelength, matedit_name, matedit_length)) {
@@ -486,65 +598,70 @@ void purge_all_vars() {
     vars_count = 0;
 }
 
-int vars_exist(int real, int cpx, int matrix) {
+bool vars_exist(int section) {
     int i;
     for (i = 0; i < vars_count; i++) {
         if ((vars[i].flags & (VAR_HIDDEN | VAR_PRIVATE)) != 0)
             continue;
+        if (section == -1)
+            return true;
         switch (vars[i].value->type) {
             case TYPE_REAL:
             case TYPE_STRING:
-                if (real)
-                    return 1;
+                if (section == CATSECT_REAL)
+                    return true;
                 else
                     break;
             case TYPE_COMPLEX:
-                if (cpx)
-                    return 1;
+                if (section == CATSECT_CPX)
+                    return true;
                 else
                     break;
             case TYPE_REALMATRIX:
             case TYPE_COMPLEXMATRIX:
-                if (matrix)
-                    return 1;
+                if (section == CATSECT_MAT)
+                    return true;
                 else
                     break;
         }
     }
-    return 0;
+    return false;
 }
 
-int contains_no_strings(const vartype_realmatrix *rm) {
+bool contains_strings(const vartype_realmatrix *rm) {
     int4 size = rm->rows * rm->columns;
-    int4 i;
-    for (i = 0; i < size; i++)
-        if (rm->array->is_string[i])
-            return 0;
-    return 1;
+    for (int4 i = 0; i < size; i++)
+        if (rm->array->is_string[i] != 0)
+            return true;
+    return false;
 }
 
+/* This is only used by core_linalg1, and does not deal with strings,
+ * even when copying a real matrix to a real matrix. It returns an
+ * error if any are encountered.
+ */
 int matrix_copy(vartype *dst, const vartype *src) {
-    int4 size, i;
     if (src->type == TYPE_REALMATRIX) {
         vartype_realmatrix *s = (vartype_realmatrix *) src;
         if (dst->type == TYPE_REALMATRIX) {
             vartype_realmatrix *d = (vartype_realmatrix *) dst;
             if (s->rows != d->rows || s->columns != d->columns)
                 return ERR_DIMENSION_ERROR;
-            size = s->rows * s->columns;
-            for (i = 0; i < size; i++) {
-                d->array->is_string[i] = s->array->is_string[i];
-                d->array->data[i] = s->array->data[i];
-            }
+            if (contains_strings(s))
+                return ERR_ALPHA_DATA_IS_INVALID;
+            int4 size = s->rows * s->columns;
+            free_long_strings(d->array->is_string, d->array->data, size);
+            memset(d->array->is_string, 0, size);
+            memcpy(d->array->data, s->array->data, size * sizeof(phloat));
             return ERR_NONE;
         } else if (dst->type == TYPE_COMPLEXMATRIX) {
             vartype_complexmatrix *d = (vartype_complexmatrix *) dst;
             if (s->rows != d->rows || s->columns != d->columns)
                 return ERR_DIMENSION_ERROR;
-            if (!contains_no_strings(s))
+            if (contains_strings(s))
                 return ERR_ALPHA_DATA_IS_INVALID;
-            size = s->rows * s->columns;
-            for (i = 0; i < size; i++) {
+            int4 size = s->rows * s->columns;
+            for (int4 i = 0; i < size; i++) {
                 d->array->data[2 * i] = s->array->data[i];
                 d->array->data[2 * i + 1] = 0;
             }
@@ -557,9 +674,8 @@ int matrix_copy(vartype *dst, const vartype *src) {
         vartype_complexmatrix *d = (vartype_complexmatrix *) dst;
         if (s->rows != d->rows || s->columns != d->columns)
             return ERR_DIMENSION_ERROR;
-        size = s->rows * s->columns * 2;
-        for (i = 0; i < size; i++)
-            d->array->data[i] = s->array->data[i];
+        int4 size = s->rows * s->columns * 2;
+        memcpy(d->array->data, s->array->data, size * sizeof(phloat));
         return ERR_NONE;
     } else
         return ERR_INVALID_TYPE;
