@@ -1,6 +1,6 @@
 /*****************************************************************************
  * Free42 -- an HP-42S calculator simulator
- * Copyright (C) 2004-2021  Thomas Okken
+ * Copyright (C) 2004-2022  Thomas Okken
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -57,6 +57,7 @@ void shell_blitter(const char *bits, int bytesperline, int x, int y,
                              int width, int height);
 
 /* shell_beeper()
+ *
  * Callback invoked by the emulator core to play a sound.
  * The first parameter is the frequency in Hz; the second is the
  * duration in ms. The sound volume is up to the GUI to control.
@@ -66,6 +67,7 @@ void shell_blitter(const char *bits, int bytesperline, int x, int y,
 void shell_beeper(int frequency, int duration);
 
 /* shell_annunciators()
+ *
  * Callback invoked by the emulator core to change the state of the display
  * annunciators (up/down, shift, print, run, battery, (g)rad).
  * Every parameter can have values 0 (turn off), 1 (turn on), or -1 (leave
@@ -86,15 +88,19 @@ void shell_annunciators(int updn, int shf, int prt, int run, int g, int rad);
  * immediately (with a return value of 1, to indicate that it would like to get
  * the CPU back as soon as possible).
  */
-int shell_wants_cpu();
+bool shell_wants_cpu();
 
-/* Callback to suspend execution for the given number of milliseconds. No event
+/* shell_delay()
+ *
+ * Callback to suspend execution for the given number of milliseconds. No event
  * processing will take place during the wait, so the core can call this
  * without having to worry about core_keydown() etc. being re-entered.
  */
 void shell_delay(int duration);
 
-/* Callback to ask the shell to call core_timeout3() after the given number of
+/* shell_request_timeout3()
+ *
+ * Callback to ask the shell to call core_timeout3() after the given number of
  * milliseconds. If there are keystroke events during that time, the timeout is
  * cancelled. (Pressing 'shift' does not cancel the timeout.)
  * This function supports the delay after SHOW, MEM, and shift-VARMENU.
@@ -102,17 +108,20 @@ void shell_delay(int duration);
 void shell_request_timeout3(int delay);
 
 /* shell_get_mem()
+ *
  * Callback to get the amount of free memory in bytes.
  */
 uint4 shell_get_mem();
 
 /* shell_low_battery()
+ *
  * Callback to find out if the battery is low. Used to emulate flag 49 and the
  * battery annunciator.
  */
-int shell_low_battery();
+bool shell_low_battery();
 
 /* shell_powerdown()
+ *
  * Callback to tell the shell that the emulator wants to power down.
  * Only called in response to OFF (shift-EXIT or the OFF command); automatic
  * power-off is left to the OS and/or shell.
@@ -120,6 +129,7 @@ int shell_low_battery();
 void shell_powerdown();
 
 /* shell_random_seed()
+ *
  * When SEED is invoked with X = 0, the random number generator should be
  * seeded to a random value; the emulator core calls this function to obtain
  * it. The core uses the least significant 14 decimal digits of the provided
@@ -129,6 +139,7 @@ void shell_powerdown();
 int8 shell_random_seed();
 
 /* shell_milliseconds()
+ *
  * Returns an elapsed-time value in milliseconds. The caller should make no
  * assumptions as to what this value is relative to; it is only intended to
  * allow the emulator core make short-term elapsed-time measurements.
@@ -136,17 +147,38 @@ int8 shell_random_seed();
 uint4 shell_milliseconds();
 
 /* shell_decimal_point()
+ *
  * Returns 0 if the host's locale uses comma as the decimal separator;
  * returns 1 if it uses dot or anything else.
  * Used to initialize flag 28 on hard reset.
  */
-int shell_decimal_point();
+bool shell_decimal_point();
+
+/* shell_date_format()
+ *
+ * Returns 0 if the host's locale uses MDY date format;
+ * returns 1 if it uses DMY;
+ * returns 2 if it uses YMD.
+ * If the host's date format doesn't match any of these three component
+ * orders, returns 0.
+ * Used to initialize flags 31 and 67 on hard reset.
+ */
+int shell_date_format();
+
+/* shell_clk24()
+ *
+ * Returns 0 if the host's locale uses a 12-hour clock
+ * returns 1 if it uses a 24-hour clock
+ * Used to initialize CLK12/CLK24 mode on hard reset.
+ */
+bool shell_clk24();
 
 /* shell_print()
+ *
  * Printer emulation. The first 2 parameters are the plain text version of the
  * data to be printed; the remaining 6 parameters are the bitmap version. The
  * former is used for text-mode copying and for spooling to text files; the
- * latter is used for graphics-mode coopying, spooling to image files, and
+ * latter is used for graphics-mode copying, spooling to image files, and
  * on-screen display.
  */
 void shell_print(const char *text, int length,
@@ -169,7 +201,7 @@ void shell_print(const char *text, int length,
  * established international standards, the one exception being acceleration,
  * which it expresses in units of Earth gravities rather than the standard
  * m/s^2. TODO: what is the exact conversion factor used by the iPhone?
- * 
+ *
  * shell_get_acceleration: x, y, z in g's (see above). Looking at the device
  * in portrait orientation, positive x points to the right, positive y points
  * up, and positive z points toward the user.
@@ -184,10 +216,10 @@ void shell_print(const char *text, int length,
  * "normalization"? I hope they mean "clipped", because otherwise you wouldn't
  * have a unit, yet they claim the unit is microteslas.
  */
-int shell_get_acceleration(double *x, double *y, double *z);
-int shell_get_location(double *lat, double *lon, double *lat_lon_acc,
+bool shell_get_acceleration(double *x, double *y, double *z);
+bool shell_get_location(double *lat, double *lon, double *lat_lon_acc,
                                 double *elev, double *elev_acc);
-int shell_get_heading(double *mag_heading, double *true_heading, double *acc,
+bool shell_get_heading(double *mag_heading, double *true_heading, double *acc,
                                 double *x, double *y, double *z);
 #endif
 
@@ -196,7 +228,7 @@ int shell_get_heading(double *mag_heading, double *true_heading, double *acc,
  * Sets and queries the "always on" state (flag 44). Pass 0 to clear, 1 to set,
  * or -1 to leave unchanged; returns the previous state.
  */
-int shell_always_on(int always_on);
+bool shell_always_on(int always_on);
 
 /* shell_get_time_date()
  *
@@ -207,7 +239,7 @@ int shell_always_on(int always_on);
 void shell_get_time_date(uint4 *time, uint4 *date, int *weekday);
 
 /* shell_message()
- * 
+ *
  * Displays a modal pop-up message box.
  */
 void shell_message(const char *message);
