@@ -1,6 +1,6 @@
 ///////////////////////////////////////////////////////////////////////////////
 // Free42 -- an HP-42S calculator simulator
-// Copyright (C) 2004-2021  Thomas Okken
+// Copyright (C) 2004-2022  Thomas Okken
 //
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License, version 2,
@@ -23,9 +23,9 @@
 void skin_menu_update(GtkWidget *w);
 void skin_load(int *width, int *height);
 
-typedef struct {
+struct SkinColor {
     unsigned char r, g, b, pad;
-} SkinColor;
+};
 
 #define IMGTYPE_MONO 1
 #define IMGTYPE_GRAY 2
